@@ -11,7 +11,7 @@ class Entry:
         semicol_index = content.find(';')
         if semicol_index >= 0:
             self.content = content[:semicol_index].strip().split()
-            self.comment = ' ' + content[semicol_index:]
+            self.comment = ' ' + content[semicol_index:].rstrip()
         else:
             self.content = content.strip().split()
             self.comment = ''
@@ -33,9 +33,15 @@ class Entry:
             nf = len(str(flt).split('.')[1])
         except IndexError:
             nf = 3
+        if nf > 12:
+            flt = float(str(flt).split('.')[0] + '.' + str(flt).split('.')[1][:12])
+            nf = len(str(flt).split('.')[1])
         if nf > dpmax:
-            nf = dpmax
-        return "{:>" + str(fields) + "." + str(nf) + "f}"
+            dpmax = nf
+            fields = dpmax + 3 + len(str(int(flt)))
+        else:
+            dpmax = nf
+        return "{:>" + str(fields) + "." + str(dpmax) + "f}"
 
     @staticmethod
     def infer_type(val):
@@ -58,6 +64,11 @@ class Entry:
     
     def __getitem__(self, item):
         return self.content[item]
+
+    def is_header(self):
+        if len(self.content) == 0:
+            return False
+        return True if (self.content[0].strip().startswith('[') and self.content[-1].strip().endswith(']')) else False
     
     def __str__(self):
         """
@@ -74,21 +85,21 @@ class EntryBonded(Entry):
     between specific atoms in the topology
     """
     fstr_suff = {('bonds', '1'): (float, float),
+                 ('bonds', '2'): (float, float),
                  ('angles', '1'): (float, float),
                  ('angles', '2'): (float, float),
                  ('angles', '5'): (float, float, float, float),
                  ('angles', '10'): (float, float),
                  ('dihedrals', '9'): (float, float, int),
                  ('dihedrals', '4'): (float, float, int),
-                 ('impropers', '4'): (float, float, int),
                  ('dihedrals', '1'): (float, float, int),
                  ('dihedrals', '3'): (float, float, float, float, float, float),
                  ('dihedrals', '2'): (float, float),
-                 ('impropers', '2'): (float, float),
-                 ('impropers', '1'): (float, float, int),  # temporary hack, fix this pls
                  ('cmap', '1'): (float,),
                  ('position_restraints', '1'): (float, float, float),
                  ('dihedral_restraints', '1'): (float, float, float),
+                 ('constraints', '2'): (float,),
+                 ('virtual_sites2', '1'): (float,),
                  ('settles', '1'): (float, float)}
 
     def __init__(self, content, subsection):
@@ -196,7 +207,7 @@ class EntryBonded(Entry):
                         fmt_suff = fmt_suff + "{:>15s} "
         fstring = self.fstring + fmt_suff
         return fstring.format(*self.atom_numbers, self.interaction_type, *self.params_state_a, *self.params_state_b) \
-            + ' ' + self.comment
+            + ' ' + self.comment + '\n'
 
         
 class EntryParam(Entry):
@@ -303,7 +314,7 @@ class EntryParam(Entry):
             try:
                 return self.format().format(*self.types, self.interaction_type, *self.modifiers, *self.params) +\
                        ' ' + self.comment + '\n'
-            except IndexError:
+            except:
                 print((*self.types, self.interaction_type, *self.modifiers, *self.params))
         else:
             return super().__str__()
@@ -350,7 +361,10 @@ class EntryAtom(Entry):
 
     def _get_atomtype_entry(self):
         atomtypes = self.subsection.section.top.parameters.get_subsection('atomtypes')
-        return [e for e in atomtypes if isinstance(e, EntryParam) and e.types[0] == self.type][-1]
+        try:
+            return [e for e in atomtypes if isinstance(e, EntryParam) and e.types[0] == self.type][-1]
+        except IndexError:
+            raise RuntimeError(f"Couldn't find non-bonded parameters for atomtype {self.type}")
 
     def __str__(self):
         fstring = self.fstring + self.float_fmt(self.charge) + self.float_fmt(self.mass) + '   '
@@ -358,7 +372,7 @@ class EntryAtom(Entry):
             alch_fstring = "{:>11s}" + self.float_fmt(self.charge_b) + self.float_fmt(self.mass_b)
             fstring += alch_fstring
             return fstring.format(self.num, self.type, self.resid, self.resname, self.atomname, self.num,
-                                  self.charge, self.mass, self.type_b, self.charge_b, self.mass_b) + self.comment
+                                  self.charge, self.mass, self.type_b, self.charge_b, self.mass_b) + self.comment + '\n'
         else:
             return fstring.format(self.num, self.type, self.resid, self.resname, self.atomname, self.num,
-                                  self.charge, self.mass) + ' ' + self.comment
+                                  self.charge, self.mass) + ' ' + self.comment + '\n'
