@@ -1,5 +1,6 @@
 import gromologist as gml
 from collections import defaultdict
+from typing import Optional, Union, Iterable
 import math
 
 
@@ -15,7 +16,8 @@ class Pdb:
                 'A': "A", 'G': "G", 'C': "C", 'U': "U", 'A5': "A", 'G5': "G", 'C5': "C", 'U5': "U",
                 'A3': "A", 'G3': "G", 'C3': "C", 'U3': "U"}
 
-    def __init__(self, filename=None, top=None, altloc='A', qt=False, **kwargs):
+    def __init__(self, filename: str = None, top: Union[str, "gml.Top"] = None, altloc: str = 'A', qt: bool = False,
+                 **kwargs):
         """
         Initializes a Pdb instnace
         :param filename: str, name of the file (.gro or .pdb)
@@ -47,10 +49,10 @@ class Pdb:
         self._atom_format_gro = "{:>5d}{:5s}{:>5s}{:>5d}{:8.3f}{:8.3f}{:8.3f}\n"
         self._cryst_format = "CRYST1{:9.3f}{:9.3f}{:9.3f}{:7.2f}{:7.2f}{:7.2f} P 1           1\n"
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return "PDB file {} with {} atoms".format(self.fname, len(self.atoms))
 
-    def add_top(self, top, **kwargs):
+    def add_top(self, top: str, **kwargs):
         """
         Adds a Top object to the current Pdb object, enabling some
         operations that couple the two
@@ -62,8 +64,16 @@ class Pdb:
         if self.top and not self.top.pdb:
             self.top.pdb = self
 
+    @property
+    def natoms(self) -> int:
+        return len(self.atoms)
+
+    @property
+    def chains(self) -> list:
+        return sorted({a.chain for a in self.atoms})
+
     @classmethod
-    def from_selection(cls, pdb, selection):
+    def from_selection(cls, pdb: "gml.Pdb", selection: str) -> "gml.Pdb":
         """
         Creates a new Pdb instance as a subset of an existing one,
         given a selection that defines a subset of atoms
@@ -81,7 +91,7 @@ class Pdb:
         return new_pdb
 
     @classmethod
-    def from_text(cls, text, orig_pdb):
+    def from_text(cls, text: str, orig_pdb: "gml.Pdb") -> "gml.Pdb":
         """
         Reads coordinates from text, useful for reading trajectories
         :param text: str, contents of the pdb/gro
@@ -104,7 +114,7 @@ class Pdb:
         new_pdb.altloc = orig_pdb.altloc
         return new_pdb
 
-    def print_protein_sequence(self, gaps=False):
+    def print_protein_sequence(self, gaps: bool = False) -> list:
         """
         Prints protein sequence chain by chain, recognizing amino acids
         as residues that contain a CA atom; unrecognized residues are written as X
@@ -114,8 +124,8 @@ class Pdb:
         chains = list({a.chain.strip() for a in self.atoms if a.atomname == 'CA'})
         sequences = []
         if not chains:
-            print("No chains in the molecule, run Pdb.add_chains() first")
-            return
+            print("No protein chains in the molecule, run Pdb.add_chains() first")
+            return []
         for ch in sorted(chains):
             cas = set(self.get_atom_indices(f'name CA and chain {ch}'))
             atoms = [a for n, a in enumerate(self.atoms) if n in cas and a.altloc in [' ', self.altloc]]
@@ -136,7 +146,7 @@ class Pdb:
                 sequences.append(''.join(seq))
         return sequences
 
-    def print_nucleic_sequence(self):
+    def print_nucleic_sequence(self) -> list:
         """
         Prints nucleic acid sequences chain by chain, recognizing nucleotides
         as residues that contain an O4' atom; unrecognized residues are written as X
@@ -147,8 +157,8 @@ class Pdb:
         mapping.update(Pdb.nucl_map)
         chains = list({a.chain.strip() for a in self.atoms if a.atomname == "O4'"})
         if not chains:
-            print("No chains in the molecule, run Pdb.add_chains() first")
-            return
+            print("No nucleic acid chains in the molecule, run Pdb.add_chains() first")
+            return []
         for ch in sorted(chains):
             cas = set(self.get_atom_indices(f"name O4' and chain {ch}"))
             atoms = [a for n, a in enumerate(self.atoms) if n in cas]
@@ -197,7 +207,8 @@ class Pdb:
                 atomlist = []
             atomlist.append(at.atomname)
 
-    def add_chains(self, selection=None, chain=None, offset=0, maxwarn=100, cutoff=10, protein_only=False, nopbc=False):
+    def add_chains(self, selection: str = None, chain: str = None, offset: int = 0, maxwarn: int = 100,
+                   cutoff: float = 10, protein_only: bool = False, nopbc: bool = False):
         """
         Given a matching Top instance, adds chain identifiers to atoms
         based on the (previously verified) matching between invididual
@@ -247,10 +258,10 @@ class Pdb:
             self.check_top(maxwarn=maxwarn)
             excluded = {'SOL', 'HOH', 'TIP3', 'WAT', 'OPC', 'K', 'NA', 'CL', 'POT', 'SOD', 'NA+', 'K+', 'CLA'}
             index = 0
-            for mol_name in self.top.system.keys():
-                if mol_name.upper() not in excluded:
-                    n_mols = self.top.system[mol_name]
-                    mol = self.top.get_molecule(mol_name)
+            for mol_count in self.top.system:
+                if mol_count[0].upper() not in excluded:
+                    n_mols = mol_count[1]
+                    mol = self.top.get_molecule(mol_count[0])
                     n_atoms = mol.natoms
                     for m in range(n_mols):
                         for a in range(n_atoms):
@@ -261,7 +272,18 @@ class Pdb:
                         if base_char > 90:
                             return
 
-    def check_top(self, maxwarn=20, fix_pdb=False, fix_top=False):
+    def list_atoms(self, selection: str = None):
+        """
+        Prints atoms in the structure, optionally
+        only the subset corresponding to the selection
+        :param selection: str, selection to be printed, default is None
+        :return: None
+        """
+        atoms = self.atoms if selection is None else self.get_atoms(selection)
+        for a in atoms:
+            print(self._write_atom(a).strip())
+
+    def check_top(self, maxwarn: int = 20, fix_pdb: bool = False, fix_top: bool = False):
         """
         Checks the structure against the associated topology to identify
         and list potential mismatches in atom names
@@ -277,7 +299,6 @@ class Pdb:
         index, err = 0, 0
         self._remove_altloc()
         for atom_top, atom_pdb in zip(self.top.atoms, self.atoms):
-            index += 1
             try:
                 rtrn = self._check_mismatch(atom_top, atom_pdb, atom_top.molname)
             except IndexError:
@@ -292,11 +313,11 @@ class Pdb:
             index += 1
             if err > maxwarn > -1:
                 raise RuntimeError("Error: too many warnings; use maxwarn=N to allow for up to N exceptions,"
-                                   "or -1 to allow for any number of them")
+                                   "or maxwarn=-1 to allow for any number of them")
         print("Check passed, all names match")
 
     @staticmethod
-    def _check_mismatch(atom_entry, atom_instance, mol_name):
+    def _check_mismatch(atom_entry: "gml.EntryAtom", atom_instance: "gml.Atom", mol_name: str) -> int:
         """
         Checks a Pdb entry against a Top entry to make sure the atom names match
         :param atom_entry: EntryAtom instance from the Top
@@ -342,14 +363,14 @@ class Pdb:
                     mol = identify(a)
             else:
                 if a.resnum != res:
-                    if len(mol) >= 1 and mol == mol_list[-1][0]:
+                    if len(mol) >= 1 and mol_list and mol == mol_list[-1][0]:
                         mol_list[-1][1] += 1
                     else:
                         mol_list.append([mol, 1])
                     mol = identify(a)
                     chain = a.chain
             res = a.resnum
-        if len(mol) >= 1 and mol == mol_list[-1][0]:
+        if len(mol) >= 1 and mol_list and mol == mol_list[-1][0]:
             mol_list[-1][1] += 1
         else:
             mol_list.append([mol, 1])
@@ -364,7 +385,7 @@ class Pdb:
         """
         self.atoms = [a for a in self.atoms if a.altloc in [' ', self.altloc]]
 
-    def _write_atom(self, atom, pdb=True):
+    def _write_atom(self, atom: "gml.Atom", pdb: bool = True) -> str:
         """
         Fn to convert an Atom instance to a line compatible with .pdb or .gro formatting
         :param atom: an Atom instance to be written
@@ -386,10 +407,10 @@ class Pdb:
                                                 atom.y / 10, atom.z / 10)
 
     @staticmethod
-    def _write_conect(atom, bonded):
+    def _write_conect(atom: int, bonded: list) -> str:
         return 'CONECT' + ('{:>5}' * (len(bonded) + 1)).format(atom, *bonded) + '\n'
 
-    def tip3_to_opc(self, offset=0.147722363):
+    def tip3_to_opc(self, offset: float = 0.147722363):
         """
         Converts a 3-point water model in a structure to a 4-point
         by adding a virtual site to each water molecule
@@ -426,7 +447,7 @@ class Pdb:
                 new_atoms.append(a)
         self.atoms = new_atoms
 
-    def renumber_atoms(self, offset=1, selection=None):
+    def renumber_atoms(self, offset: int = 1, selection: Optional[str] = None):
         """
         Consecutively renumbers all atoms in the structure
         :param offset: int, number of the first atom to be set
@@ -437,27 +458,32 @@ class Pdb:
         for n, atom in enumerate(ats, offset):
             atom.serial = n
 
-    def renumber_residues(self, offset=1, selection=None):
+    def renumber_residues(self, offset: int = 1, selection: Optional[str] = None, per_chain: bool = False):
         """
         Consecutively renumbers all residues in the structure
         :param offset: int, number of the first residue to be set
         :param selection: str, will only renumber residues matching the selection
+        :param per_chain: bool, whether numbering of each chain should start from `offset`
         :return: None
         """
         count = offset
         ats = range(len(self.atoms)) if selection is None else self.get_atom_indices(selection)
         for n in ats:
+            if per_chain and n > 1 and self.atoms[n].chain != self.atoms[n-1].chain:
+                count = offset
             temp = count
             try:
                 if self.atoms[n].resnum != self.atoms[n + 1].resnum or self.atoms[n].chain != self.atoms[n + 1].chain:
                     # avoid resnum==10000 as it gets written as 0 and causes problems:
                     temp = count + 1 if not (count + 1 % 10000 == 0) else count + 2
+                    # TODO fix the above to work with residue numbers like -1, -2 etc.
             except IndexError:
                 pass
             self.atoms[n].resnum = count
             count = temp
 
-    def reposition_atom_from_hook(self, atomsel, hooksel, bondlength, p1_sel=None, p2_sel=None, vector=None):
+    def reposition_atom_from_hook(self, atomsel: str, hooksel: str, bondlength: float, p1_sel: Optional[str] = None,
+                                  p2_sel: Optional[str] = None, vector: Optional[Iterable] = None):
         """
         Sets coordinates of an atom based on a "hook" atom (one it's bound to), bond length, and a vector defining
         the direction of the bond
@@ -484,7 +510,7 @@ class Pdb:
         scale = bondlength / vec_len
         movable.set_coords([h + scale * v for h, v in zip(hook_xyz, vec)])
 
-    def match_order_by_top_names(self, arange=None):
+    def match_order_by_top_names(self, arange: Optional[tuple] = None):
         """
         Whenever PDB atoms have different ordering than .top ones
         but naming is consistent, we can use the ordering from .top
@@ -500,14 +526,15 @@ class Pdb:
             raise ValueError("a Top object has not been assigned; molecule info missing")
         new_atoms = []
         index = 0
-        for mol_name in self.top.system.keys():
-            mol = self.top.get_molecule(mol_name)
-            n_mols = self.top.system[mol_name]
+        # TODO is arange here working at all?
+        for mol_count in self.top.system:
+            mol = self.top.get_molecule(mol_count[0])
+            n_mols = mol_count[1]
             atom_subsection = mol.get_subsection('atoms')
             atom_entries = [e for e in atom_subsection if isinstance(e, gml.EntryAtom)]
-            for _ in arange(n_mols):
+            for _ in range(n_mols):
                 for a in atom_entries:
-                    if not arange or arange[0] <= index < arange[1]:
+                    if arange is None or arange[0] <= index < arange[1]:
                         pdb_loc = self.get_atom_indices("resname {} and resid {} and name {}".format(a.resname, a.resid,
                                                                                                      a.atomname))
                         pdb_loc = [loc for loc in pdb_loc if not arange or arange[0] <= index < arange[1]]
@@ -521,8 +548,9 @@ class Pdb:
                     index += 1
         self.atoms = new_atoms
 
-    def insert_atom(self, serial, name="HX", hooksel=None, bondlength=None, p1_sel=None, p2_sel=None,
-                    vector=None, renumber=True, **kwargs):
+    def insert_atom(self, serial: int, name: str = "HX", hooksel: Optional[str] = None,
+                    bondlength: Optional[float] = None, p1_sel: Optional[str] = None, p2_sel: Optional[str] = None,
+                    vector: Optional[Iterable] = None, renumber: Optional[bool] = True, **kwargs):
         """
         Inserts an atom into the atomlist. The atom is defined by
         providing a base Atom instance and a number of keyworded modifiers,
@@ -540,6 +568,7 @@ class Pdb:
         """
         base_atom = self.get_atom(hooksel)
         new_atom = Atom(self._write_atom(base_atom))
+        kwargs.update({"atomname": name})
         for kw in kwargs.keys():
             if kw not in {"serial", "atomname", "resname", "chain", "resnum", "x", "y", "z", "occ", "beta", "element"}:
                 raise ValueError("{} is not a valid Atom attribute")
@@ -547,12 +576,13 @@ class Pdb:
         self.atoms.insert(serial-1, new_atom)
         if renumber:
             self.renumber_atoms()
-        atomsel = f'name {name} and resid {base_atom.resnum} and resname {base_atom.resname}'
+        rname = base_atom.resname if "resname" not in kwargs.keys() else kwargs["resname"]
+        atomsel = f'name {name} and resid {base_atom.resnum} and resname {rname}'
         if (atomsel is not None and hooksel is not None and bondlength is not None and
                 ((p1_sel is not None and p2_sel is not None) or vector is not None)):
             self.reposition_atom_from_hook(atomsel, hooksel, bondlength, p1_sel, p2_sel, vector)
 
-    def delete_atom(self, serial, renumber=False):
+    def delete_atom(self, serial: int, renumber: bool = False):
         """
         Removes an atom from the Pdb object
         :param serial: int, serial number of the atom to be removed (usually 1-based)
@@ -579,7 +609,7 @@ class Pdb:
             if not a.element:
                 a.element = [x for x in a.atomname if not x.isdigit()][0]
 
-    def get_atom_indices(self, selection_string):
+    def get_atom_indices(self, selection_string: str) -> list:
         """
         Applies a selection to the structure and returns the 0-based indices of selected atoms
         :param selection_string: str, consistent with the selection language syntax (see README)
@@ -641,17 +671,37 @@ class Pdb:
         :return: set, a broadened list of atom indices
         """
         new_list = []
+        use_numpy = False
+        dist_array = None
+        try:
+            import numpy as np
+        except ImportError:
+            pass
+        else:
+            use_numpy = True
+            dist_array = np.array(self.get_coords())
         for n, atom in enumerate(self.atoms):
             if nopbc:
-                if any([self._atoms_dist(atom, self.atoms[query]) <= threshold for query in query_iter]):
-                    new_list.append(n)
+                if not use_numpy:
+                    if any([self._atoms_dist(atom, self.atoms[query]) <= threshold for query in query_iter]):
+                        new_list.append(n)
             else:
                 if any([self._atoms_dist_pbc(atom, self.atoms[query]) <= threshold for query in query_iter]):
                     new_list.append(n)
+        if nopbc and use_numpy:
+            for n in query_iter:
+                vecs = np.linalg.norm(dist_array - dist_array[n, :], axis=1)
+                new_list.extend(list(np.where(vecs <= threshold)[0]))
         return set(new_list)
 
     @staticmethod
     def _atoms_dist(at1, at2):
+        """
+        Calculates the (non-PBC-corrected) distance between atoms at1 and at2
+        :param at1: Atom, 1st atom defining the distance
+        :param at2: Atom, 2nd atom defining the distance
+        :return: float, distance
+        """
         return ((at2.x - at1.x) ** 2 + (at2.y - at1.y) ** 2 + (at2.z - at1.z) ** 2) ** 0.5
 
     def _atoms_dist_pbc(self, at1, at2):
@@ -744,12 +794,23 @@ class Pdb:
         atoms_remove = mutant.atoms_to_remove()
         for at in atoms_remove:
             equivalents = {'OG': 'OG1', 'HG': 'HG1', 'HG1': 'HG11', 'HG2': 'HG12', 'HG3': 'HG13', 'CG': 'CG1',
-                           'CD': 'CD1', 'HD': 'HD1', 'HD1': 'HD11', 'HD2': 'HD12', 'HD3': 'HD13'}
+                           'CD': 'CD1', 'HD': 'HD1', 'HD1': ['HD11', 'HE2'], 'HD2': 'HD12', 'HD3': 'HD13'}
             print("Removing atom {} from resid {} in structure".format(at, resid))
+            atnum = None
             try:
                 atnum = self.get_atom_index('{}resid {} and name {}'.format(chstr, resid, at))
             except RuntimeError:
-                atnum = self.get_atom_index('{}resid {} and name {}'.format(chstr, resid, equivalents[at]))
+                equivs = equivalents[at] if isinstance(equivalents[at], list) else [equivalents[at]]
+                for equiv in equivs:
+                    try:
+                        atnum = self.get_atom_index('{}resid {} and name {}'.format(chstr, resid, equiv))
+                    except RuntimeError:
+                        continue
+                    else:
+                        break
+                else:
+                    if atnum is None:
+                        raise RuntimeError(f"Couldn't find any of the following: {at}, {', '.join(equivs)}")
             _ = self.atoms.pop(atnum)
         for atom_add, hook, geo_ref, bond_length, aft in zip(atoms_add, hooks, geo_refs, bond_lengths, afters):
             print("Adding atom {} to resid {} in structure".format(atom_add, resid))
@@ -813,7 +874,8 @@ class Pdb:
         f = -1 if nv == 3 else 1
         return f * sum(v[0] for v in vecs) / nv, f * sum(v[1] for v in vecs) / nv, f * sum(v[2] for v in vecs) / nv
 
-    def add_vs2(self, resid, name1, name2, vsname='V1', fraction=0.5):
+    def add_vs2(self, resid: int, name1: str, name2: str, vsname: str = 'V1', fraction: float = 0.5,
+                serial: Optional[int] = None, chain: Optional[str] = None):
         """
         Adds a virtual site (VS) defined by two atoms, interpolating between
         their coordinates
@@ -822,15 +884,18 @@ class Pdb:
         :param name2: str, 2nd atom to use for the construction of the VS
         :param vsname: str, name of the new virtual site
         :param fraction: float, where to put the VS (0 = on atom 1, 1 = on atom 2, can be interpolated or extrapolated)
+        :param serial: int, where to locate the new VS in the atomlist
+        :param chain: str, additional specification to identify the atoms
         :return: None
         """
-        serial = self.get_atoms(f"resid {resid}")[-1].serial + 2
-        a1 = self.get_atom(f"resid {resid} and name {name1}")
-        a2 = self.get_atom(f"resid {resid} and name {name2}")
+        chsel = f' and chain {chain}' if chain is not None else ''
+        serial = self.get_atoms(f"resid {resid}{chsel}")[-1].serial + 2 if serial is None else serial
+        a1 = self.get_atom(f"resid {resid} and name {name1}{chsel}")
+        a2 = self.get_atom(f"resid {resid} and name {name2}{chsel}")
         dist = self._atoms_dist_pbc(a1, a2)
-        self.insert_atom(serial, name=vsname, hooksel=f"resid {resid} and name {name1}",
-                         bondlength=dist * fraction, p1_sel=f"resid {resid} and name {name1}",
-                         p2_sel=f"resid {resid} and name {name2}", atomname=vsname)
+        self.insert_atom(serial, name=vsname, hooksel=f"resid {resid} and name {name1}{chsel}",
+                         bondlength=dist * fraction, p1_sel=f"resid {resid} and name {name1}{chsel}",
+                         p2_sel=f"resid {resid} and name {name2}{chsel}", atomname=vsname)
 
     def interatomic_dist(self, resid1=1, resid2=2):
         """
@@ -883,10 +948,11 @@ class Pdb:
             n = self.get_atom(f'name {at1} and resnum {resnum} and resname {resname} {chn}')
             c = self.get_atom(f'name {at2} and resnum {resnum} and resname {resname} {chn}')
             h = self.get_atom(f'name {at3} and resnum {resnum} and resname {resname} {chn}')
-            chi = self._get_chirality(at, n, c, h, nopbc)
-            if chi < -2.5 or 0 > chi > -1.5:
+            chi = self._get_chirality([n, c, h, at], nopbc)
+            if chi < -0.9 or 0 > chi > -0.35:
                 if printing:
                     print(f"Check {label} for residue {resname} num {resnum}, looks a bit off")
+                    print(chi)
                 else:
                     return False
             elif chi > 0:
@@ -912,25 +978,25 @@ class Pdb:
         """
         self.reposition_atom_from_hook(h_sel, c_sel, 1.09, h_sel, c_sel)
 
-    def _get_chirality(self, *atoms, nopbc=False):
+    def _get_chirality(self, atomlist, nopbc=False):
         """
         Calculates a dihedral defined by 4 atoms that
         constitute a chiral center
-        :param atoms: list of len 4, atoms defining the chiral center
+        :param atomlist: list of len 4, atoms defining the chiral center
         :param nopbc: bool, whether to ignore PBC when calculating vectors
         :return: float, the dihedral's value
         """
         if nopbc:
-            v1 = self._atoms_vec(atoms[1], atoms[0])
-            v2 = self._atoms_vec(atoms[0], atoms[2])
-            v3 = self._atoms_vec(atoms[2], atoms[3])
+            v1 = self._atoms_vec(atomlist[0], atomlist[1])
+            v2 = self._atoms_vec(atomlist[1], atomlist[2])
+            v3 = self._atoms_vec(atomlist[2], atomlist[3])
         else:
-            v1 = self._atoms_vec_pbc(atoms[1], atoms[0])
-            v2 = self._atoms_vec_pbc(atoms[0], atoms[2])
-            v3 = self._atoms_vec_pbc(atoms[2], atoms[3])
+            v1 = self._atoms_vec_pbc(atomlist[0], atomlist[1])
+            v2 = self._atoms_vec_pbc(atomlist[1], atomlist[2])
+            v3 = self._atoms_vec_pbc(atomlist[2], atomlist[3])
         n1 = self._normalize(self._cross_product(v1, v2))
         n2 = self._normalize(self._cross_product(v2, v3))
-        m1 = self._cross_product(n1, self._normalize(v2))
+        m1 = self._normalize(self._cross_product(n1, self._normalize(v2)))
         x = self._scalar_product(n1, n2)
         y = self._scalar_product(m1, n2)
         return math.atan2(y, x)
@@ -999,7 +1065,7 @@ class Pdb:
             raise RuntimeError('Can\'t read box properties')
         return atoms, tuple(box), remarks
 
-    def add_conect(self, cutoff=1.55):
+    def add_conect(self, cutoff=1.65):
         """
         Adds CONECT entries to the PDB, using the value of
         cutoff to determine molecule connectivity
@@ -1009,14 +1075,16 @@ class Pdb:
         try:
             import numpy as np
         except ImportError:
-            # TODO implement a numpy-free version
-            raise RuntimeError("This functionality requires numpy to be present")
+            for n, atom in enumerate(self.atoms):
+                dists = [self._atoms_dist(atom, at) for at in self.atoms]
+                selected = [self.atoms[m].serial for m, a in enumerate(dists) if 0.1 < a < cutoff]
+                self.conect[atom.serial] = selected
         else:
             coords = np.array(self.get_coords())
             for n, atom in enumerate(self.atoms):
                 dists = np.linalg.norm(coords - np.array(atom.coords), axis=1)
-                selected = list(np.where(np.logical_and(dists < cutoff, dists > 0.1))[0] + 1)
-                self.conect[n + 1] = selected
+                selected = [self.atoms[i].serial for i in list(np.where(np.logical_and(dists < cutoff, dists > 0.1))[0])]
+                self.conect[atom.serial] = selected
 
     def set_beta(self, values, selection=None, smooth=False, ignore_mem=False):
         """
