@@ -1,6 +1,8 @@
 import os
+import platform
 import datetime
 from copy import deepcopy
+from typing import Optional, Iterable, TextIO, Union
 
 import gromologist as gml
 from collections import OrderedDict
@@ -20,15 +22,18 @@ class Top:
         """
         # TODO maybe allow for construction of a blank top with a possibility to read data later?
         self.suppress = suppress
-        self.gromacs_dir, self.gmx_exe = self.find_gmx_dir()
+        self.gromacs_dir, self.gmx_exe = gml.find_gmx_dir(suppress)
         self.gromacs_dir = gmx_dir if not self.gromacs_dir else self.gromacs_dir
         self.gmx_exe = gmx_exe if not self.gmx_exe else self.gmx_exe
         self.pdb = None
         self.rtp = {}
         self.pdb = None if pdb is None else gml.Pdb(pdb, top=self)
         self.fname = filename
-        self.top = self.fname.split('/')[-1]
-        self.dir = os.getcwd() + '/' + '/'.join(self.fname.split('/')[:-1])
+        self.top = self.fname.split(os.sep)[-1]
+        if self.fname.startswith(os.sep):
+            self.dir = os.sep.join(self.fname.split(os.sep)[:-1])
+        else:
+            self.dir = os.getcwd() + os.sep + os.sep.join(self.fname.split(os.sep)[:-1])
         with open(self.fname) as top_file:
             self._contents = top_file.readlines()
         self.defines = {}
@@ -46,72 +51,126 @@ class Top:
         self._parse_sections()
 
     @property
-    def system(self):
+    def system(self) -> list:
+        """
+        Returns a list of tuples with the contents of the system,
+        following the [ molecules ] section of a topology
+        """
         if not self.top.endswith('top'):
-            return {}
+            return []
         else:
             return self.read_system_properties()
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return "Topology with {} atoms and total charge {:.3f}".format(self.natoms, self.charge)
 
     def print(self, *args):
+        """
+        A custom Print function that can be turned out with suppress=True
+        """
         if not self.suppress:
             print(*args)
 
     @classmethod
-    def _from_text(cls, text, gmx_dir=None, pdb=None, ignore_ifdef=False):
+    def _from_text(cls, text: str, gmx_dir: Optional[str] = None, pdb: Optional[str] = None, ignore_ifdef: bool = False):
+        """
+        A simple wrapper to generate a topology from an in-memory string object
+        :param text: str, the text to be parsed
+        :param gmx_dir: str, optional path to the gromacs directory
+        :param pdb: str, optional path to the corresponding structure
+        :param ignore_ifdef: bool, optional to ignore missing #ifdefs
+        :return: a Top instance
+        """
         with open('tmp_topfile.gromo', 'w') as tmp:
             tmp.write(text)
         instance = cls('tmp_topfile.gromo', gmx_dir, pdb, ignore_ifdef)
         os.remove('tmp_topfile.gromo')
         return instance
 
-    def find_gmx_dir(self):
-        """
-        Attempts to find Gromacs internal files to fall back to
-        when default .itp files are included using the
-        #include statement
-        :return: str, path to share/gromacs/top directory
-        """
-        gmx = os.popen('which gmx 2> /dev/null').read().strip()
-        if not gmx:
-            gmx = os.popen('which gmx_mpi 2> /dev/null').read().strip()
-        if not gmx:
-            gmx = os.popen('which gmx_d 2> /dev/null').read().strip()
-        if gmx:
-            gmx_dir = '/'.join(gmx.split('/')[:-2]) + '/share/gromacs/top'
-            self.print('Gromacs files found in directory {}'.format(gmx_dir))
-            return gmx_dir, gmx
-        else:
-            self.print('No working Gromacs compilation found, assuming all file dependencies are referred to locally; '
-                       'to change this, make the gmx executable visible in $PATH or specify gmx_dir for the Topology')
-            return False, False
-
     @property
-    def molecules(self):
+    def molecules(self) -> list:
+        """
+        A property attribute returning the list of all SectionMols corresponding to molecules
+        :return: list of SectionMol objects
+        """
         return [s for s in self.sections if isinstance(s, gml.SectionMol)]
 
     @property
-    def alchemical_molecules(self):
+    def alchemical_molecules(self) -> list:
+        """
+        As Top.molecules but only returns alchemical molecules (ones that have state B defined)
+        :return: list of SectionMol objects
+        """
         return [s for s in self.sections if isinstance(s, gml.SectionMol) and s.is_alchemical]
 
     @property
-    def parameters(self):
+    def parameters(self) -> "gml.SectionParam":
+        """
+        A property attribute that returns the SectionParams section containing all parameter sets
+        :return: list
+        """
         return [s for s in self.sections if isinstance(s, gml.SectionParam)][0]
 
     @property
-    def atoms(self):
+    def atoms(self) -> list:
+        """
+        A property attribute that returns a list of all atoms in the system
+        :return: list of EntryAtom entries
+        """
         atomlist = []
-        for mol in self.system.keys():
-            mol_sect = [s for s in self.molecules if s.mol_name == mol][0]
-            for q in range(self.system[mol]):
-                for a in mol_sect.atoms:
+        for mol_count in self.system:
+            molecule = self.get_molecule(mol_count[0])
+            for q in range(mol_count[1]):
+                for a in molecule.atoms:
                     atomlist.append(a)
         return atomlist
 
+    def remove_molecule(self, molname: str):
+        """
+        Removes a molecule definition and the corresponding entry from the system
+        definition (e.g. SOL to remove the solvent)
+        :param molname: str, name of the molecule to remove
+        :return: None
+        """
+        section = [s for s in self.sections if isinstance(s, gml.SectionMol) and s.mol_name == molname][0]
+        self.sections.remove(section)
+        system_subsection = [s.get_subsection('molecules') for s in self.sections if
+                             'molecules' in [ss.header for ss in s.subsections]][0]
+        entries_to_del = [e for e in system_subsection.entries if str(e).split()[0] == molname]
+        for entry in entries_to_del:
+            system_subsection.entries.remove(entry)
+
+    def select_atoms(self, selection_string: str) -> list:
+        """
+        Returns atoms' indices according to the specified selection string
+        :param selection_string: str, a VMD-compatible selection
+        :return: list, 0-based indices of atoms compatible with the selection
+        """
+        sel = gml.SelectionParser(self)
+        return sel(selection_string)
+
+    def select_atom(self, selection_string: str) -> int:
+        """
+        Returns atoms' indices according to the specified selection string
+        :param selection_string: str, a VMD-compatible selection
+        :return: int, 0-based index of atom compatible with the selection
+        """
+        sel = gml.SelectionParser(self)
+        result = sel(selection_string)
+        if len(result) > 1:
+            raise RuntimeError("Selection {} returned more than one atom: {}".format(selection_string, result))
+        elif len(result) < 1:
+            raise RuntimeError("Selection {} returned no atoms".format(selection_string, result))
+        return result[0]
+
+    def get_atoms(self, selection_string: str) -> list:
+        return [self.atoms[i] for i in self.select_atoms(selection_string)]
+
+    def get_atom(self, selection_string: str):
+        return self.atoms[self.select_atom(selection_string)]
+
     @property
-    def defined_atomtypes(self):
+    def defined_atomtypes(self) -> set:
         return {ent.types[0] for ent in self.parameters.get_subsection('atomtypes').entries_param}
 
     def list_molecules(self):
@@ -119,25 +178,28 @@ class Top:
         Prints out a list of molecules contained in the System
         :return: None
         """
-        for mol in self.system.keys():
-            print("{:20s}{:>10d}".format(mol, self.system[mol]))
+        for mol_count in self.system:
+            print("{:20s}{:>10d}".format(mol_count[0], mol_count[1]))
 
-    def clear_ff_params(self, section='all'):
+    def clear_ff_params(self, section: str = 'all'):
         used_params = []
         for mol in self.molecules:
             used_params.extend(mol.find_used_ff_params(section=section))
         used_params.extend(self.parameters.find_used_ff_params(section=section))
         self.parameters.clean_unused(used_params, section=section)
 
-    def add_pdb(self, pdbfile):
+    def add_pdb(self, pdbfile: Union[str, "gml.Pdb"]):
         """
         Allows to pair a PDB file with the topology after the instance was initialized
-        :param pdbfile: str, path to PDB file
+        :param pdbfile: str, path to PDB file, or gml.Pdb object
         :return: None
         """
-        self.pdb = gml.Pdb(pdbfile, top=self)
+        if isinstance(pdbfile, str):
+            self.pdb = gml.Pdb(pdbfile, top=self)
+        else:
+            self.pdb = pdbfile
 
-    def add_ff_params(self, section='all'):
+    def add_ff_params(self, section: str = 'all'):
         """
         Explicitly puts FF parameters in sections 'bonds', 'angles',
         'dihedrals' so that the resulting topology is independent of
@@ -148,8 +210,79 @@ class Top:
         for mol in self.molecules:
             mol.add_ff_params(add_section=section)
 
-    def find_missing_ff_params(self, section='all', fix_by_analogy=False, fix_B_from_A=False, fix_A_from_B=False,
-                               once=False):
+    def add_molecule_from_file(self, filename: str, molnames: Optional[list] = None):
+        """
+        Adds a molecule from an external file (can be .itp or .top) to the current topology
+        :param filename: name of the file containing the molecule to be added
+        :param molnames: list, enumerates molecules to be added (can be just 1-element list), None means add all
+        :return: None
+        """
+        contents = [line for line in open(filename)]
+        special_sections = {'defaults', 'moleculetype', 'system'}
+        special_lines = [n for n, l in enumerate(contents)
+                         if l.strip() and l.strip().strip('[]').strip().split()[0] in special_sections]
+        special_lines.append(len(contents))
+        for beg, end in zip(special_lines[:-1], special_lines[1:]):
+            if 'moleculetype' in contents[beg]:
+                molsections = [n for n, i in enumerate(self.sections) if isinstance(i, gml.SectionMol)][-1]
+                section = self._yield_sec(contents[beg:end])
+                if molnames is None or section.mol_name in molnames:
+                    self.sections.insert(molsections+1, section)
+        print("Molecules inserted. Try running Top.find_missing_ff_params() to see if the topology contains"
+              "all necessary parameters.\n\nTo add the newly defined molecule to the system, use "
+              "Top.add_molecules_to_system() or manually edit the [ molecules ] section in the topology")
+
+    def add_parameters_from_file(self, filename: str, sections: Optional[list] = None, overwrite: bool = False):
+        """
+        Adds parameters from an external file
+        :param filename: name of the file containing the parameters to be added
+        :param sections: list, enumerates sections to be added (can be just 1-element list), None means add all
+        :param overwrite: bool, whether to overwrite existing parameters in case of conflict (default is not)
+        :return: None
+        """
+        other = gml.Top(filename, ignore_ifdef=True)
+        if not all([i == j for i, j in zip(self.parameters.get_subsection('defaults').entries_param[0].content,
+                                           other.parameters.get_subsection('defaults').entries_param[0].content)]):
+            raise RuntimeError("Can't merge parameters with different [ defaults ] sections, "
+                               "make sure they are identical")
+        for subsection in other.parameters.subsections:  # TODO make optional selections available
+            if subsection.header != 'defaults':
+                try:
+                    # let's check if we already have this subsection in our topo
+                    own_subs = self.parameters.get_subsection(subsection.header)
+                except KeyError:
+                    # if not, let's add it as it is assuming it's non-empty
+                    if len(subsection.entries_param) > 0:
+                        self.parameters.subsections.append()
+                else:
+                    own_subs._combine_entries(subsection, overwrite)
+
+    def add_molecules_to_system(self, molname: str, nmol: int):
+        """
+        Adds a specified number of molecules to the system (specified
+        in the [ molecules ] section at the end of the .top file)
+        :param molname: str, name of the molecule (has to be already defined in the topology)
+        :param nmol: int, number of molecules to be repeated
+        :return: None
+        """
+        mollist = [mol.mol_name for mol in self.molecules]
+        if molname not in mollist:
+            raise RuntimeError(f"Molecule {molname} not found among defined molecules ({mollist}), please add it"
+                               f"manually or via Top.add_molecule_from_itp()")
+        system_subsection = [s.get_subsection('molecules') for s in self.sections
+                             if 'molecules' in [ss.header for ss in s.subsections]]
+        if len(system_subsection) > 1:
+            raise RuntimeError("Multiple 'molecules' subsection found in the topology, this is not allowed")
+        elif len(system_subsection) == 0:
+            self.print("Section 'molecules' not present in the topology, will be created now")
+            system_subsection = self._yield_sec(["[ molecules ]"])
+            self.sections.append(system_subsection)
+        else:
+            system_subsection = system_subsection[0]
+        system_subsection.add_entry(gml.Entry(f"{molname} {nmol}", system_subsection))
+
+    def find_missing_ff_params(self, section: str = 'all', fix_by_analogy: bool = False, fix_B_from_A: bool = False,
+                               fix_A_from_B: bool = False, fix_dummy: bool = False, once: bool = False):
         """
         Identifies FF parameters that are not defined in sections
         'bondtypes', angletypes', ...; if required, will attempt to
@@ -158,18 +291,20 @@ class Top:
         :param fix_by_analogy: dict, if set, will attempt to use params by analogy, matching key types to value types
         :param fix_B_from_A: bool, will assign params for state B from state A
         :param fix_A_from_B: bool, will assign params for state A from state B
+        :param fix_dummy: bool, will assign zeros as parameters
         :param once: bool, will only print a given missing term once per molecule
         :return: None
         """
+        # TODO check if all types are defined
         for mol in self.molecules:
-            mol.find_missing_ff_params(section, fix_by_analogy, fix_B_from_A, fix_A_from_B, once=once)
+            mol.find_missing_ff_params(section, fix_by_analogy, fix_B_from_A, fix_A_from_B, fix_dummy, once=once)
 
-    def add_posres(self, keyword='POSRES', value=1000):
+    def add_posres(self, keyword: str = 'POSRES', value: int = 1000):
         for mol in self.molecules:
             if len(mol.atoms) > 3:
                 mol.add_posres(keyword, value)
 
-    def add_params_file(self, paramfile):
+    def add_params_file(self, paramfile: str):
         prmtop = Top._from_text('#include {}\n'.format(paramfile))
         own_defentry = [e for e in self.sections[0].subsections[0].entries if isinstance(e, gml.EntryParam)][0]
         other_defentry = [e for e in prmtop.sections[0].subsections[0].entries if isinstance(e, gml.EntryParam)][0]
@@ -183,7 +318,7 @@ class Top:
         paramsect_own.subsections.extend(paramsect_other.subsections)
         paramsect_own._merge()
 
-    def _include_all(self, ign_ifdef):
+    def _include_all(self, ign_ifdef: bool):
         """
         includes all .itp files in the .top file to facilitate processing
         :return: None
@@ -201,7 +336,7 @@ class Top:
             lines = [i for i in range(len(self._contents)) if self._contents[i].startswith("#include")
                      and i not in ignore_lines]
 
-    def _preprocess_conditional_includes(self, ifdef):
+    def _preprocess_conditional_includes(self, ifdef: list):
         """
         Because of the implementation of CHARMM36m/CHARMM36 in gmx,
         the two variants are chosen depending on a preprocessing conditional
@@ -230,7 +365,7 @@ class Top:
                 self._contents.pop(lnum)
             self._contents.insert(start_final[0], incl)
 
-    def _find_ifdef_lines(self):
+    def _find_ifdef_lines(self) -> set:
         """
         Finds #ifdef/#endif blocks in the topology if user
         explicitly asks to ignore them
@@ -247,7 +382,7 @@ class Top:
                 ignore_set.add(n)
         return ignore_set
 
-    def resolve_ifdefs(self, ifdefs):
+    def resolve_ifdefs(self, ifdefs: list):
         continuing = True
         while continuing:
             remove_lines = []
@@ -319,17 +454,17 @@ class Top:
             raise AttributeError("System properties have not been read, this is likely not a complete .top file")
         sections_to_delete = []
         for section_num, section in enumerate(self.sections):
-            if isinstance(section, gml.SectionMol) and section.mol_name not in self.system.keys():
+            if isinstance(section, gml.SectionMol) and section.mol_name not in {x[0] for x in self.system}:
                 sections_to_delete.append(section_num)
         self.sections = [s for n, s in enumerate(self.sections) if n not in sections_to_delete]
 
-    def _find_in_path(self, filename):
+    def _find_in_path(self, filename: str) -> (str, str):
         """
         looks for a file to be included in either the current directory
         or in Gromacs directories (as given by user), in order to
         include all .itp files in a single .top file
         :param filename: str, name of the file to be searched for
-        :return: None
+        :return: str, full path to the file, path to the file directory
         """
         if filename.strip().startswith('./'):
             filename = filename.strip()[2:]
@@ -347,7 +482,7 @@ class Top:
                                     ' ignore_ifdef=True'.format(filename))
 
     @staticmethod
-    def _add_prefix_to_include(content, prefix):
+    def _add_prefix_to_include(content: list, prefix: str) -> list:
         """
         Modifies #include statements if nested #includes
         point to different directories
@@ -378,13 +513,15 @@ class Top:
         special_lines.append(len(self._contents))
         for beg, end in zip(special_lines[:-1], special_lines[1:]):
             self.sections.append(self._yield_sec(self._contents[beg:end]))
+            excess_if = self.count_ifs(beg) - self.count_ifs(beg, endif=True)
+            self.sections[-1].conditional = excess_if
         # in case there are #defines at the very beginning (e.g. CHARMM36):
         for lnum in range(0, special_lines[0]):
             if not self._contents[lnum].lstrip().startswith(';') and self._contents[lnum].strip():
                 entry = gml.Entry(self._contents[lnum].strip(), self.sections[0].subsections[0])
                 self.header.append(entry)
 
-    def _yield_sec(self, content):
+    def _yield_sec(self, content: list) -> "gml.Section":
         """
         Chooses which class (Section or derived classes)
         should be used for the particular set of entries
@@ -398,31 +535,47 @@ class Top:
         else:
             return gml.Section(content, self)
 
-    def read_system_properties(self):
+    def count_ifs(self, linenum: int, endif: bool = False) -> int:
+        """
+        Counts #if or #endif directives up to line linenum
+        :param linenum: int, line number
+        :param endif: bool, if True we're looking for #endif instead of #if
+        :return: int, number of directives found
+        """
+        if not endif:
+            return len([ln for ln in self._contents[:linenum] if
+                        ln.strip().startswith("#ifdef") or ln.strip().startswith("#ifndef")])
+        else:
+            return len([ln for ln in self._contents[:linenum] if ln.strip().startswith("#endif")])
+
+    def read_system_properties(self) -> list:
         """
         Reads in system composition based on the [ molecules ] section
-        :return: None
+        :return: list of tuples (Mol_name, number_of_molecules)
         """
         system_subsection = [s.get_subsection('molecules') for s in self.sections
                              if 'molecules' in [ss.header for ss in s.subsections]]
-        molecules = OrderedDict()  # we want to preserve the order of molecules in the system for e.g. PDB checking
+        molecules = []  # we want to preserve the order of molecules in the system for e.g. PDB checking
         if len(system_subsection) > 1:
             raise RuntimeError("Multiple 'molecules' subsection found in the topology, this is not allowed")
         elif len(system_subsection) == 0:
             self.print("Section 'molecules' not present in the topology, assuming this is an isolated .itp")
-            return None
+            return []
         for e in system_subsection[0]:
             if e.content:
-                molecules[e.content[0]] = int(e.content[1])
+                molecules.append((e.content[0], int(e.content[1])))
         return molecules
 
     @property
-    def charge(self):
-        return sum([self.system[mol] * self.get_molecule(mol).charge for mol in self.system.keys()])
+    def charge(self) -> float:
+        return sum([mol_count[1] * self.get_molecule(mol_count[0]).charge for mol_count in self.system])
 
     @property
-    def natoms(self):
-        return sum([self.system[mol] * self.get_molecule(mol).natoms for mol in self.system.keys()])
+    def natoms(self) -> int:
+        return sum([mol_count[1] * self.get_molecule(mol_count[0]).natoms for mol_count in self.system])
+
+    def nmol(self, name: Optional[str] = None) -> int:
+        return sum([x[1] for x in self.system if name is None or x[0] == name])
 
     def explicit_defines(self):
         """
@@ -436,7 +589,7 @@ class Top:
                 if isinstance(s, gml.SubsectionBonded):
                     s.explicit_defines()
 
-    def get_molecule(self, mol_name):
+    def get_molecule(self, mol_name: str) -> "gml.SectionMol":
         """
         Finds a molecule (SectionMol instance) whose mol_name
         matches the query name
@@ -450,20 +603,23 @@ class Top:
             raise RuntimeError("Molecule {} is duplicated in topology".format(mol_name))
         return mol[0]
 
-    def check_pdb(self, maxwarn=None):
+    def check_pdb(self, maxwarn: Optional[int] = None, fix_pdb: bool = False, fix_top: bool = False):
         """c2r.gro
         Compares the topology with a PDB object to check
         for consistency, just as gmx grompp does;
         if inconsistencies are found, prints a report
+        :param maxwarn: int, maximum number of warnings to print, default is 20
+        :param fix_pdb: bool, whether to set names in Pdb using names from the Top
+        :param fix_top: bool, whether to set names in Top using names from the Pdb
         :return: None
         """
         if self.pdb:
-            mw = 20 if maxwarn is None else int(maxwarn)
-            self.pdb.check_top(mw)
+            mw = 20 if maxwarn is None else maxwarn
+            self.pdb.check_top(mw, fix_pdb=fix_pdb, fix_top=fix_top)
         else:
             raise AttributeError("No PDB file has been bound to this topology")
 
-    def save_top(self, outname='merged.top', split=False):
+    def save_top(self, outname: str = 'merged.top', split: bool = False):
         """
         Saves the combined topology to the specified file
         :param outname: str, file name for output
@@ -521,7 +677,13 @@ class Top:
         for mol in self.molecules:
             mol.recalc_qtot()
 
-    def solute_tempering(self, temperatures, molecules):
+    def solute_tempering(self, temperatures: list, molecules: list):
+        """
+        Prepares .top files for REST2
+        :param temperatures: list of float, set of "fake" temperatures for REST2 (lowest should be first)
+        :param molecules: list of int, indices of molecules that will have their parameters modified
+        :return: None
+        """
         self.explicit_defines()
         for n, t in enumerate(temperatures):
             self.print(f'generating topology for effective temperature of {t} K...')
@@ -532,19 +694,30 @@ class Top:
             mod.save_top(self.fname.replace('.top', f'-rest{temperatures[0]/t:.3f}.top'))
 
     @staticmethod
-    def _write_section(outfile, section):
+    def _write_section(outfile: TextIO, section):
+        """
+        Writes a single section to the output file
+        :param outfile: an open file in/out object, links to the output file
+        :param section: a Section object to be written
+        :return:
+        """
         for subsection in section.subsections:
-            outfile.write('\n[ {} ]\n'.format(subsection.write_header))
+            outfile.write('\n[ {} ]\n'.format(subsection.header))
             for entry in subsection:
                 str_entry = str(entry).rstrip() + '\n'
                 outfile.write(str_entry)
 
-    def _write_header(self, outfile):
+    def _write_header(self, outfile: TextIO):
+        """
+        Writes the header to a newly generated file
+        :param outfile: an open file in/out object, links to the output file
+        :return: None
+        """
         outname = outfile.name.split('/')[-1]
         outfile.write(";\n;  File {} was generated with the gromologist library\n"
                       ";  by user: {}\n;  on host: {}\n;  at date: {} \n;\n".format(outname,
-                                                                                    os.getenv("USER"),
-                                                                                    os.uname()[1],
+                                                                                    platform.os.getenv("USER"),
+                                                                                    platform.uname()[1],
                                                                                     datetime.datetime.now()))
         for entry in self.header:
             str_entry = str(entry).rstrip() + '\n'
